@@ -11,6 +11,7 @@
 from future.moves.collections import OrderedDict
 
 import errno
+import json
 import logging
 import os
 import stat
@@ -20,12 +21,28 @@ import boto3
 import configparser
 from botocore.exceptions import ClientError
 
-from pcluster.config.mappings import ALIASES, AWS, CLUSTER, GLOBAL
-from pcluster.utils import get_instance_vcpus, get_stack, get_stack_name
+from pcluster.cluster_model import ClusterModel, get_cluster_model, infer_cluster_model
+from pcluster.config.cfn_param_types import ClusterCfnSection
+from pcluster.config.mappings import ALIASES, AWS, GLOBAL
+from pcluster.config.param_types import StorageData
+from pcluster.utils import (
+    get_cfn_param,
+    get_file_section_name,
+    get_installed_version,
+    get_stack,
+    get_stack_name,
+    get_stack_version,
+    is_hit_enabled_cluster,
+)
 
 LOGGER = logging.getLogger(__name__)
 
 
+def default_config_file_path():
+    """Return the default path for the ParallelCluster configuration file."""
+    return os.path.expanduser(os.path.join("~", ".parallelcluster", "config"))
+
+
 class PclusterConfig(object):
     """
     Class to manage the configuration of a cluster created (or to create) with ParallelCluster.
@@ -40,6 +57,9 @@ class PclusterConfig(object):
         fail_on_file_absence=False,
         fail_on_error=None,
         cluster_name=None,
+        auto_refresh=True,
+        enforce_version=True,
+        skip_load_json_config=False,
     ):
         """
         Initialize object, from file, from a CFN Stack or from the internal mapping.
@@ -55,9 +75,17 @@ class PclusterConfig(object):
         # "From Stack" initialization parameters:
         :param cluster_name: the cluster name associated to a running Stack,
         if specified the initialization will start from the running Stack
+        :param auto_refresh: if set, refresh() method will be called every time something changes in the structure of
+        the configuration, like a section being added, removed or renamed.
+        :param enforce_version: when True enforces the CLI version to be of the same version as the cluster the user
+        is interacting with.
         """
+        self.__autorefresh = False  # Initialization in progress
         self.fail_on_error = fail_on_error
-        self.sections = OrderedDict({})
+        self.cfn_stack = None
+        self.__sections = OrderedDict({})
+        self.__enforce_version = enforce_version
+        self.__skip_load_json_config = skip_load_json_config
 
         # always parse the configuration file if there, to get AWS section
         self._init_config_parser(config_file, fail_on_file_absence)
@@ -68,10 +96,16 @@ class PclusterConfig(object):
 
         # init pcluster_config object, from cfn or from config_file
         if cluster_name:
+            self.cluster_name = cluster_name
             self.__init_sections_from_cfn(cluster_name)
         else:
             self.__init_sections_from_file(cluster_label, self.config_parser, fail_on_file_absence)
 
+        self.__autorefresh = auto_refresh  # Initialization completed
+
+        # Refresh sections and parameters
+        self._config_updated()
+
     def _init_config_parser(self, config_file, fail_on_config_file_absence=True):
         """
         Parse the config file and initialize config_file and config_parser attributes.
@@ -86,7 +120,7 @@ class PclusterConfig(object):
             self.config_file = os.environ["AWS_PCLUSTER_CONFIG_FILE"]
             default_config = False
         else:
-            self.config_file = os.path.expanduser(os.path.join("~", ".parallelcluster", "config"))
+            self.config_file = default_config_file_path()
             default_config = True
 
         self.config_file = str(self.config_file)
@@ -110,6 +144,29 @@ class PclusterConfig(object):
         except (configparser.ParsingError, configparser.DuplicateOptionError) as e:
             self.error("Error parsing configuration file {0}.\n{1}".format(self.config_file, str(e)))
 
+    @staticmethod
+    def get_global_section_keys():
+        """Return the keys associated to the global sections, not related to the cluster one."""
+        return ["aws", "aliases", "global"]
+
+    def get_section_keys(self, include_global_sections=False, excluded_keys=None):
+        """Return the section keys."""
+        excluded_keys = excluded_keys or []
+        if not include_global_sections:
+            excluded_keys += self.get_global_section_keys()
+
+        section_keys = [section_key for section_key in self.__sections.keys() if section_key not in excluded_keys]
+        return section_keys
+
+    def _get_file_section_names(self):
+        """Return the names of the sections as represented in the configuration file."""
+        file_section_names = []
+        for section_key, sections in self.__sections.items():
+            for _, section in sections.items():
+                file_section_names.append(get_file_section_name(section_key, section.label))
+
+        return file_section_names
+
     def get_sections(self, section_key):
         """
         Get the Section(s) identified by the given key.
@@ -124,7 +181,7 @@ class PclusterConfig(object):
         :param section_key: the identifier of the section type
         :return a dictionary containing the section
         """
-        return self.sections.get(section_key, {})
+        return self.__sections.get(section_key, {})
 
     def get_section(self, section_key, section_label=None):
         """
@@ -157,21 +214,33 @@ class PclusterConfig(object):
 
         :param section, a Section object
         """
-        if section.key not in self.sections:
-            self.sections[section.key] = {}
+        if section.key not in self.__sections:
+            self.__sections[section.key] = OrderedDict({})
 
         section_label = section.label if section.label else section.definition.get("default_label", "default")
-        self.sections[section.key][section_label] = section
+        self.__sections[section.key][section_label] = section
+        self._config_updated()
 
-    def remove_section(self, section_key, section_label):
+    def remove_section(self, section_key, section_label=None):
         """
         Remove a section from the PclusterConfig object, if there.
 
         :param section_key: the identifier of the section type
         :param section_label: the label of the section to delete.
         """
-        if section_key in self.sections:
-            self.sections[section_key].pop(section_label, None)
+        if section_key in self.__sections:
+            sections = self.__sections[section_key]
+
+            if section_label:
+                # If section label is specified, remove it directly
+                sections.pop(section_label)
+            else:
+                # If no label is specified, check that no more than one section exists with the provided key
+                if len(sections) > 1:
+                    raise Exception("More than one section with key {0}".format(section_key))
+                else:
+                    self.__sections.pop(section_key)
+        self._config_updated()
 
     def __init_aws_credentials(self):
         """Set credentials in the environment to be available for all the boto3 calls."""
@@ -190,6 +259,15 @@ class PclusterConfig(object):
             # we rely on the AWS CLI configuration or already set env variable
             pass
 
+    @property
+    def cluster_model(self):
+        """Get the cluster model used in the configuration."""
+        cluster_model = ClusterModel.SIT
+        cluster_section = self.get_section("cluster")
+        if cluster_section:
+            cluster_model = get_cluster_model(cluster_section.definition.get("cluster_model"))
+        return cluster_model
+
     @property
     def region(self):
         """Get the region. The value is stored inside the aws_region_name of the aws section."""
@@ -227,29 +305,41 @@ class PclusterConfig(object):
         """Set fail_on_error property value."""
         self._fail_on_error = fail_on_error
 
-    def to_file(self):
+    def to_file(self, print_stdout=False, exclude_unrelated_sections=False):
         """Convert the internal representation of the cluster to the relative file sections."""
-        for section_key in ["aws", "global", "aliases"]:
+        if exclude_unrelated_sections:
+            # Remove sections not strictly related to the cluster from the config_parser.
+            cluster_related_sections = self._get_file_section_names()
+            for section_name in self.config_parser.sections():
+                if section_name not in cluster_related_sections:
+                    self.config_parser.remove_section(section_name)
+
+        for section_key in self.get_global_section_keys():
             self.get_section(section_key).to_file(self.config_parser, write_defaults=True)
 
         self.get_section("cluster").to_file(self.config_parser)
 
-        # ensure that the directory for the config file exists
-        if not os.path.isfile(self.config_file):
-            try:
-                config_folder = os.path.dirname(self.config_file) or "."
-                os.makedirs(config_folder)
-            except OSError as e:
-                if e.errno != errno.EEXIST:
-                    raise  # can safely ignore EEXISTS for this purpose...
+        if print_stdout:
+            # print log file to stdout instead of writing the file
+            self.config_parser.write(sys.stdout)
+            sys.stdout.flush()
+        else:
+            # ensure that the directory for the config file exists
+            if not os.path.isfile(self.config_file):
+                try:
+                    config_folder = os.path.dirname(self.config_file) or "."
+                    os.makedirs(config_folder)
+                except OSError as e:
+                    if e.errno != errno.EEXIST:
+                        raise  # can safely ignore EEXISTS for this purpose...
 
-            # Fix permissions
-            with open(self.config_file, "a"):
-                os.chmod(self.config_file, stat.S_IRUSR | stat.S_IWUSR)
+                # Fix permissions
+                with open(self.config_file, "a"):
+                    os.chmod(self.config_file, stat.S_IRUSR | stat.S_IWUSR)
 
-        # Write configuration to disk
-        with open(self.config_file, "w") as conf_file_stream:
-            self.config_parser.write(conf_file_stream)
+            # Write configuration to disk
+            with open(self.config_file, "w") as conf_file_stream:
+                self.config_parser.write(conf_file_stream)
 
     def to_cfn(self):
         """
@@ -257,7 +347,18 @@ class PclusterConfig(object):
 
         :return: a dict containing the cfn parameters associated with the cluster configuration
         """
-        return self.get_section("cluster").to_cfn()
+        return self.to_storage().cfn_params
+
+    def to_storage(self):
+        """
+        Get a data structure with all the information needed to persist the configuration.
+
+        The internal representation of the cluster is converted into a data structure containing the information to be
+        stored into all the storage mechanisms used by the CLI (currently CloudFormation parameters and Json).
+
+        :return: a dict containing the cfn parameters and the json dict associated with the cluster configuration
+        """
+        return self.get_section("cluster").to_storage()
 
     def __init_sections_from_file(self, cluster_label=None, config_parser=None, fail_on_absence=False):
         """
@@ -275,8 +376,15 @@ class PclusterConfig(object):
             cluster_label = (
                 self.get_section("global").get_param_value("cluster_template") if self.get_section("global") else None
             )
+
+        # Infer cluster model and load cluster section accordingly
+        cluster_model = infer_cluster_model(config_parser=config_parser, cluster_label=cluster_label)
+
         self.__init_section_from_file(
-            CLUSTER, config_parser, section_label=cluster_label, fail_on_absence=fail_on_absence
+            cluster_model.get_cluster_section_definition(),
+            config_parser,
+            section_label=cluster_label,
+            fail_on_absence=fail_on_absence,
         )
 
     def __init_section_from_file(self, section_definition, config_parser, section_label=None, fail_on_absence=False):
@@ -296,15 +404,73 @@ class PclusterConfig(object):
         except configparser.NoSectionError as e:
             self.error("Section '[{0}]' not found in the config file.".format(e.section))
 
+    @property
+    def auto_refresh(self):
+        """Return the configuration autorefresh."""
+        return self.__autorefresh
+
+    @auto_refresh.setter
+    def auto_refresh(self, refresh_enabled):
+        """Enable or disable the configuration autorefresh."""
+        self.__autorefresh = refresh_enabled
+
+    def _config_updated(self):
+        """
+        Notify the PclusterConfig instance that the configuration structure has changed.
+
+        The purpose of this method is to allow internal configuration objects such as Param, Section etc to notify the
+        parent PclusterConfig when something structural has changed. The configuration will be reloaded based on whether
+        or not the autofresh function is enabled.
+        """
+        if self.__autorefresh:
+            self.refresh()
+
+    def refresh(self):
+        """
+        Reload the sections structure and refresh all configuration sections and parameters.
+
+        This method must be called if structural configuration changes have been applied, like updating a section
+        label, adding or removing a section etc.
+        """
+        # Rebuild the new sections structure
+        new_sections = OrderedDict({})
+        for key, sections in self.__sections.items():
+            new_sections_map = OrderedDict({})
+            for _, section in sections.items():
+                new_sections_map[section.label] = section
+            new_sections[key] = new_sections_map
+        self.__sections = new_sections
+
+        # Refresh all sections
+        for _, sections in self.__sections.items():
+            for _, section in sections.items():
+                section.refresh()
+
     def __init_sections_from_cfn(self, cluster_name):
         try:
-            stack = get_stack(get_stack_name(cluster_name))
+            self.cfn_stack = get_stack(get_stack_name(cluster_name))
+            if self.__enforce_version and get_stack_version(self.cfn_stack) != get_installed_version():
+                self.error(
+                    "The cluster {0} was created with a different version of ParallelCluster: {1}. "
+                    "Installed version is {2}. This operation may only be performed using the same ParallelCluster "
+                    "version used to create the cluster.".format(
+                        cluster_name, get_stack_version(self.cfn_stack), get_installed_version()
+                    )
+                )
 
-            section_type = CLUSTER.get("type")
-            section = section_type(section_definition=CLUSTER, pcluster_config=self).from_cfn_params(
-                cfn_params=stack.get("Parameters", [])
+            cfn_params = self.cfn_stack.get("Parameters")
+            json_params = self.__load_json_config(self.cfn_stack) if not self.__skip_load_json_config else None
+
+            # Infer cluster model and load cluster section accordingly
+            cluster_model = infer_cluster_model(cfn_stack=self.cfn_stack)
+            section = ClusterCfnSection(
+                section_definition=cluster_model.get_cluster_section_definition(), pcluster_config=self
             )
+
             self.add_section(section)
+
+            section.from_storage(StorageData(cfn_params, json_params))
+
         except ClientError as e:
             self.error(
                 "Unable to retrieve the configuration of the cluster '{0}'.\n{1}".format(
@@ -314,7 +480,7 @@ class PclusterConfig(object):
 
     def validate(self):
         """Validate the configuration."""
-        for _, sections in self.sections.items():
+        for _, sections in self.__sections.items():
             for _, section in sections.items():
                 section.validate()
 
@@ -329,136 +495,47 @@ class PclusterConfig(object):
         """Get the Availability zone of the Compute Subnet."""
         return self.get_section("vpc").get_param_value("compute_availability_zone")
 
+    def __load_json_config(self, cfn_stack):
+        """Retrieve Json configuration params from the S3 bucket linked from the cfn params."""
+        json_config = None
+        if is_hit_enabled_cluster(cfn_stack):
+            s3_bucket_name = get_cfn_param(cfn_stack.get("Parameters"), "ResourcesS3Bucket")
+
+            if not s3_bucket_name or s3_bucket_name == "NONE":
+                self.error("Unable to retrieve configuration: ResourceS3Bucket not available.")
+
+            json_config = self.__retrieve_cluster_config(s3_bucket_name)
+
+        return json_config
+
+    def __retrieve_cluster_config(self, bucket):
+        table = boto3.resource("dynamodb").Table(get_stack_name(self.cluster_name))
+        config_version = None  # Use latest if not found
+        try:
+            config_version_item = table.get_item(ConsistentRead=True, Key={"Id": "CLUSTER_CONFIG"})
+            if config_version_item or "Item" in config_version_item:
+                config_version = config_version_item["Item"].get("Version")
+        except Exception as e:
+            self.error("Failed when retrieving cluster config version from DynamoDB with error {0}".format(e))
+
+        try:
+            config_version_args = {"VersionId": config_version} if config_version else {}
+            s3_object = boto3.resource("s3").Object(bucket, "configs/cluster-config.json")
+            json_str = s3_object.get(**config_version_args)["Body"].read().decode("utf-8")
+            return json.loads(json_str, object_pairs_hook=OrderedDict)
+        except Exception as e:
+            self.error("Unable to load configuration from bucket '{0}'.\n{1}".format(bucket, e))
+
     def __test_configuration(self):  # noqa: C901
         """
-        Try to launch the requested instances (in dry-run mode) to verify configuration parameters.
+        Perform global tests to verify that the wanted cluster configuration can be deployed in the user's account.
 
-        NOTE: The number of max instances is set to 1 because run_instances in dry-mode doesn't try to allocate the
-        requested instances. The goal of the test is verify the provided configuration.
+        Check operations may involve dryrun tests and/or other AWS calls and depend on the current cluster model.
         """
         LOGGER.debug("Testing configuration parameters...")
-        cluster_section = self.get_section("cluster")
-        vpc_section = self.get_section("vpc")
-
-        if (
-            not cluster_section
-            or cluster_section.get_param_value("scheduler") == "awsbatch"
-            or cluster_section.get_param_value("cluster_type") == "spot"
-            or not vpc_section
-        ):
-            return
-
-        # Retrieve instance types
-        master_instance_type = cluster_section.get_param_value("master_instance_type")
-        compute_instance_type = cluster_section.get_param_value("compute_instance_type")
-
-        # Retrieve network parameters
-        compute_subnet = vpc_section.get_param_value("compute_subnet_id")
-        master_subnet = vpc_section.get_param_value("master_subnet_id")
-        vpc_security_group = vpc_section.get_param_value("vpc_security_group_id")
-        if not compute_subnet:
-            compute_subnet = master_subnet
-        security_groups_ids = []
-        if vpc_security_group:
-            security_groups_ids.append(vpc_security_group)
-
-        # Initialize CpuOptions
-        disable_hyperthreading = cluster_section.get_param_value("disable_hyperthreading")
-        master_vcpus = get_instance_vcpus(self.region, master_instance_type)
-        compute_vcpus = get_instance_vcpus(self.region, compute_instance_type)
-        master_cpu_options = {"CoreCount": master_vcpus // 2, "ThreadsPerCore": 1} if disable_hyperthreading else {}
-        compute_cpu_options = {"CoreCount": compute_vcpus // 2, "ThreadsPerCore": 1} if disable_hyperthreading else {}
-
-        # Initialize Placement Group Logic
-        placement_group = cluster_section.get_param_value("placement_group")
-        placement = cluster_section.get_param_value("placement")
-        master_placement_group = (
-            {"GroupName": placement_group}
-            if placement_group not in [None, "NONE", "DYNAMIC"] and placement == "cluster"
-            else {}
-        )
-        compute_placement_group = (
-            {"GroupName": placement_group} if placement_group not in [None, "NONE", "DYNAMIC"] else {}
-        )
-
-        try:
-            latest_alinux_ami_id = self.__get_latest_alinux_ami_id()
-
-            # Test Master Instance Configuration
-            self.__ec2_run_instance(
-                InstanceType=master_instance_type,
-                MinCount=1,
-                MaxCount=1,
-                ImageId=latest_alinux_ami_id,
-                SubnetId=master_subnet,
-                SecurityGroupIds=security_groups_ids,
-                CpuOptions=master_cpu_options,
-                Placement=master_placement_group,
-                DryRun=True,
-            )
-
-            # Test Compute Instances Configuration
-            self.__ec2_run_instance(
-                InstanceType=compute_instance_type,
-                MinCount=1,
-                MaxCount=1,
-                ImageId=latest_alinux_ami_id,
-                SubnetId=compute_subnet,
-                SecurityGroupIds=security_groups_ids,
-                CpuOptions=compute_cpu_options,
-                Placement=compute_placement_group,
-                DryRun=True,
-            )
-        except ClientError:
-            self.error("Unable to validate configuration parameters.")
+        self.cluster_model.test_configuration(self)
         LOGGER.debug("Configuration parameters tested correctly.")
 
-    def __get_latest_alinux_ami_id(self):
-        """Get latest alinux ami id."""
-        try:
-            alinux_ami_id = (
-                boto3.client("ssm")
-                .get_parameters_by_path(Path="/aws/service/ami-amazon-linux-latest")
-                .get("Parameters")[0]
-                .get("Value")
-            )
-        except ClientError as e:
-            self.error("Unable to retrieve Amazon Linux AMI id.\n{0}".format(e.response.get("Error").get("Message")))
-            raise
-
-        return alinux_ami_id
-
-    def __ec2_run_instance(self, **kwargs):
-        """Wrap ec2 run_instance call. Useful since a successful run_instance call signals 'DryRunOperation'."""
-        try:
-            boto3.client("ec2").run_instances(**kwargs)
-        except ClientError as e:
-            code = e.response.get("Error").get("Code")
-            message = e.response.get("Error").get("Message")
-            if code == "DryRunOperation":
-                pass
-            elif code == "UnsupportedOperation":
-                if "does not support specifying CpuOptions" in message:
-                    self.error(message.replace("CpuOptions", "disable_hyperthreading"))
-                self.error(message)
-            elif code == "InstanceLimitExceeded":
-                self.error(
-                    "You've reached the limit on the number of instances you can run concurrently "
-                    "for the configured instance type.\n{0}".format(message)
-                )
-            elif code == "InsufficientInstanceCapacity":
-                self.error("There is not enough capacity to fulfill your request.\n{0}".format(message))
-            elif code == "InsufficientFreeAddressesInSubnet":
-                self.error(
-                    "The specified subnet does not contain enough free private IP addresses "
-                    "to fulfill your request.\n{0}".format(message)
-                )
-            else:
-                self.error(
-                    "Unable to validate configuration parameters. "
-                    "Please double check your cluster configuration.\n{0}".format(message)
-                )
-
     def error(self, message):
         """Print an error message and Raise SystemExit exception to the stderr if fail_on_error is true."""
         if self.fail_on_error:
@@ -479,4 +556,19 @@ class PclusterConfig(object):
         configuration settings.
         :param config_file: pcluster config file - None to use default
         """
-        PclusterConfig(config_file=config_file, fail_on_error=False, fail_on_file_absence=False)
+        PclusterConfig(config_file=config_file, fail_on_error=False, fail_on_file_absence=False, auto_refresh=False)
+
+    def update(self, pcluster_config):
+        """
+        Update the entire configuration structure taking the data from the provided pcluster_config instance.
+
+        This operation allows the configuration metadata to be correctly updated before being sent back to
+        CloudFormation.
+
+        :param pcluster_config: The new configuration containing the updated settings.
+        """
+        # When the configuration is updated, all parameters are replaced except config metadata
+        # which is needed to keep the right linking between sections and CloudFormation resources
+        config_metadata_param = self.get_section("cluster").get_param("cluster_config_metadata")
+        self.__sections = pcluster_config.__sections
+        self.get_section("cluster").set_param("cluster_config_metadata", config_metadata_param)
